package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.ClosedReceiveChannelException
import kotlinx.coroutines.channels.ClosedSendChannelException
import net.sergeych.crypto.encodeVarUnsigned
import net.sergeych.crypto.readVarUnsigned
import net.sergeych.kiloparsec.Transport
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.warning
import net.sergeych.mp_tools.globalLaunch
import java.net.StandardSocketOptions.TCP_NODELAY
import java.nio.ByteBuffer
import java.nio.channels.AsynchronousSocketChannel
import kotlin.coroutines.cancellation.CancellationException
import kotlin.coroutines.suspendCoroutine

private val log = LogTag("ASTD")

/**
 * Convert asynchronous socket to a [Transport.Device] using non-blocking nio,
 * in a coroutine-effective manner. Note that it runs coroutines to read/write
 * to the socket in a global scope.These are closed when transport is closed
 *  or the socket is closed, for example, by network failure.
 */
suspend fun asyncSocketToDevice(socket: AsynchronousSocketChannel): Transport.Device {
    val deferredDevice = CompletableDeferred<Transport.Device>()
    globalLaunch {
        coroutineScope {
            fun stop() {
                cancel()
            }
            socket.setOption(TCP_NODELAY, true)
            val input = Channel<UByte>(1024)
            val output = Channel<UByte>(1024)
            // copy from socket to input
            launch {
                val inb = ByteBuffer.allocate(1024)
                while (isActive) {
                    val size: Int = suspendCoroutine { continuation ->
                        socket.read(inb, continuation, IntCompletionHandler)
                    }
                    if (size < 0) stop()
                    else for (i in 0..<size) input.send(inb[i].toUByte().also { print(it.toInt().toChar()) })
                }
            }
            // copy from output tp socket
            launch {
                val outBuff = ArrayList<Byte>(1024)
                try {
                    while (isActive) {
                        outBuff.clear()
                        outBuff.add(output.receive().toByte())
                        while (!output.isEmpty)
                            outBuff.add(output.receive().toByte())
                        suspendCoroutine { continuation ->
                            socket.write(ByteBuffer.wrap(outBuff.toByteArray()), continuation, IntCompletionHandler)
                        }
                    }
                } catch (_: ClosedReceiveChannelException) {
                    stop()
                }
            }
            // pump blocks from socket output to device input
            val inputBlocks = Channel<UByteArray?>()
            launch {
                try {
                    while (isActive) {
                        val size = readVarUnsigned(input)
                        if (size == 0u) log.warning { "zero size block is ignored!" }
                        else {
                            val block = UByteArray(size.toInt())
                            for (i in 0..<size.toInt()) {
                                block[i] = input.receive()
                            }
                            inputBlocks.send(block)
                        }
                    }
                } catch (_: CancellationException) {
                    inputBlocks.send(null)
                } catch (_: ClosedReceiveChannelException) {
                    inputBlocks.send(null)
                    stop()
                }
            }
            val outputBlocks = Channel<UByteArray>()
            launch {
                try {
                    while (isActive) {
                        val block = outputBlocks.receive()
                        output.sendAll(encodeVarUnsigned(block.size.toUInt()))
                        output.sendAll(block)
                    }
                } catch (_: ClosedSendChannelException) {
                    stop()
                }
            }
            deferredDevice.complete(
                ProxyDevice(inputBlocks, outputBlocks) { stop() }
            )
        }
        globalLaunch { socket.close() }
    }
    return deferredDevice.await()
}
